# -*- coding: UTF-8 -*
'''
Created on 2018年11月23日

@author: Robin
'''
from __future__ import print_function

__version__ = '1.7.4'


def md5(s):
    import hashlib
    m2 = hashlib.md5()
    m2.update(s.encode("utf8"))
    return m2.hexdigest()


def get_sign(salt, key):
    salt = str(salt)
    key = str(key)
    return md5(md5(salt) + md5(key))


class XIOTException(Exception):
    pass


class App(object):
    '''用户应用'''

    def __init__(self, servurl, appid, appkey):
        '''创建需要应用'''
        self.servurl = servurl.rstrip('/')
        self.appid = appid
        self.appkey = appkey

    def doApi(self, apiname, data):
        import requests, json, time
        time = int(time.time())
        data['_appid'] = self.appid
        data['_time'] = time
        data['_sign'] = get_sign(time, self.appkey)

        url = '%s/iot/%s' % (self.servurl, apiname)
        headers = {"Content-Type": "application/json"}
        res = requests.post(url, data=json.dumps(data), headers=headers).json()
        if not res or res['code'] != 0:
            raise XIOTException('%s:%s' % (res.get('code', -1), res.get('msg', '')))
        return res


class DTU(object):
    def __init__(self, app, dtu_id):
        '''app: 应用对象  dtu_id:DTU的ID(字符串)'''
        self.app = app
        self.dtu_id = str(dtu_id)
        self.running = True

    def doApi(self, apiname, data):
        '''执行API'''
        return self.app.doApi(apiname, data)

    def updateDevices(self, devs):
        '''批量更新多个设备'''
        data = {
            'dtu_id': self.dtu_id,
            'devices': devs,
        }
        return self.doApi('update', data)

    def updateValues(self, dev_id, vals, time=None):
        '''批量更新单个设备的多个值。'''
        dev = {'dev_id': str(dev_id), 'attr': vals}
        if time != None:
            dev['time'] = time
        return self.updateDevices([dev])

    def updateValue(self, dev_id, attr_id, value, time=None):
        '''更新指定设备的指定值。dev_id:设备ID(字符串), attr_id:属性ID(字符串), value:属性值(任意Python类型)，time:时间戳(time.time()获取)'''
        return self.updateValues(dev_id, {str(attr_id): value}, time=time)

    def sendEvent(self, dev_id, evt_id, value='', time=None):
        data = {
            'dtu_id': self.dtu_id,
            'dev_id': dev_id,
            'evt_id': evt_id,
            'value': value,
        }
        if time != None:
            data['time'] = time
        return self.doApi('event', data)

    def stop(self):
        self.running = False


# MQTT
class MQTTClient(object):
    def __init__(self, host='127.0.0.1', port=1883, username=None, password=None, clientid=None):
        from paho.mqtt.client import Client
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        if not clientid:
            import uuid
            clientid = str(uuid.uuid4())
        self.mqttclient = Client(client_id=clientid, clean_session=True)
        self.mqttclient.on_connect = self._on_connect  # lambda client, userdata, flags, rc: self._on_connect(client, userdata, flags, rc)
        self.mqttclient.on_message = self._on_message  # lambda client, userdata, msg: self._on_message(client, userdata, msg)
        self.mqttclient.on_disconnect = self._on_disconnect
        self.submap = {}
        self.connected = False

    def connect(self):
        import threading
        if self.username:
            self.mqttclient.username_pw_set(self.username, self.password)
        self.mqttclient.connect(self.host, self.port)
        threading.Thread(target=self._forloop).start()

    def disconnect(self):
        self.mqttclient.disconnect()

    def _subscribe(self, topic, qos=0):
        self.mqttclient.subscribe(topic=topic, qos=qos)

    def subscribe(self, topic, qos=0):
        self.submap[topic] = {
            'topic': topic,
            'qos': qos
        }
        if self.connected:
            self._subscribe(topic=topic, qos=qos)

    def publish(self, topic, payload=None, qos=0, retain=False):
        import six
        if not isinstance(payload, six.string_types):
            import json
            payload = json.dumps(payload)
        return self.mqttclient.publish(topic, payload=payload, qos=qos, retain=retain)

    def _forloop(self):
        self.mqttclient.loop_forever()

    def _on_connect(self, client, userdata, flags, rc):
        # print("_on_connect", client, userdata, flags, rc)
        self.connected = True
        for v in self.submap.values():
            self._subscribe(topic=v['topic'], qos=v['qos'])

    def _on_disconnect(self, client, userdata, rc):
        # print("_on_disconnect", client, userdata, rc)
        # if rc != 0:
        #     print('retry')
        self.connected = False

    def _on_message(self, client, userdata, msg):
        import json
        # print("_on_message", client, userdata, msg.topic, msg.payload)
        try:
            msgd = json.loads(msg.payload.decode('utf-8'))
        except:
            print('not json', msg.topic, msg.payload)
            return
        try:
            self.on_message(msg.topic, msgd)
        except:
            import traceback
            traceback.print_exc()
            print(msg.topic, msg.payload)

    def on_message(self, topic, data):
        pass


class MQTTDTU(DTU):
    def __init__(self, app, dtu_id, mqtt_host='127.0.0.1', mqtt_port=1883, mqtt_username=None, mqtt_password=None):
        '''app: 应用对象  dtu_id:DTU的ID(字符串) mqtt_host:MQTT主机 mqtt_port:MQTT端口'''
        super(MQTTDTU, self).__init__(app, dtu_id)
        self.mqtt_host = mqtt_host
        self.mqtt_port = mqtt_port
        dtu = self

        class Client(MQTTClient):
            def on_message(self, topic, data):
                if not data or type(data) != dict:
                    return
                if not data or type(data) != dict:
                    return
                typename = data.get('type')
                datafrom = data.get('_from')
                topicls = topic.split('/')
                # print(topic, data)
                r = None
                if len(topicls) == 3:
                    appid, dtu_id, dev_id = topicls
                    if typename == 'action' and data.get('data', {}).get('action'):
                        r = dtu.on_action(dev_id, action=data['data']['action'], value=data['data'].get('value'))
                    elif typename == 'notify' and data.get('data'):
                        r = dtu.on_notify(dev_id, data=data['data'])
                if r is not None and data.get('_return'):
                    # 动作回执
                    ret = data['_return']
                    # print('r', data, r)
                    self.publish(ret['topic'], {
                        'key': ret['key'],
                        'data': r
                    })

        client = Client(host=self.mqtt_host, port=self.mqtt_port, username=mqtt_username, password=mqtt_password)
        client.connect()
        topic = '%s/%s/#' % (self.app.appid, self.dtu_id)
        client.subscribe(topic)
        self.client = client

    def doApi(self, apiname, data):
        topic = '%s/%s' % (self.app.appid, self.dtu_id)
        return self.client.publish(topic, {'type': apiname, 'data': data})

    def stop(self):
        super(MQTTDTU, self).stop()
        self.client.disconnect()

    def on_action(self, dev_id, action, value=None):
        '''执行动作'''
        print('on_action', dev_id, action, value)

    def on_notify(self, dev_id, data):
        '''通知'''
        print('on_notify', dev_id, data)

    def sendNotify(self, dev_id, data):
        '''发送通知'''
        topic = '%s/%s/%s' % (self.app.appid, self.dtu_id, dev_id)
        return self.client.publish(topic, {'type': 'notify', 'data': data})


def test():
    # 需python requests包，如果没安装的话执行: pip install requests
    import time
    # servurl = 'https://xiot.inruan.com/'
    servurl = 'http://127.0.0.1:8080/'
    app = App(servurl, 'xt15429507412810938', '775563e1cce37176cc33aa5d053d9e07')

    # 创建DTU对象
    dtu = DTU(app, 'dtu1')

    # print(dtu.updateValue('dev1', 'voltage', int(time.time()) % 50 / 10.0))
    # return

    class TestMQTTDTU(MQTTDTU):
        def on_action(self, dev_id, action, value=None):
            print(dev_id, action, value)

    # dtu = TestMQTTDTU(app, 'dtu2', mqtt_host='172.16.1.142')

    import signal
    def stop(a, b):
        print('stop')
        dtu.stop()

    signal.signal(signal.SIGINT, stop)
    import random
    while dtu.running:
        try:
            # print(dtu.updateValue('dev3', 'time', int(time.time()) - 1545700000 - 6442565))
            # print(dtu.sendEvent(dev_id='dev1', evt_id=random.choices(['normal', 'warn', 'error'])[0], value=random.choices(['电压过低设备警告', '未知故障', '紧急报警'])[0]))
            # print(dtu.updateValue('dev1', 'time', int(time.time()) - 1545700000 - 6442565))
            print(dtu.updateValue('dev1', 'status', 'error'))
            time.sleep(5)
            print(dtu.updateValue('dev1', 'status', 'connected'))
            time.sleep(5)
            print(dtu.updateValue('dev1', 'status', 'warning'))
            time.sleep(5)
        except:
            import traceback
            traceback.print_exc()
        ct = 1
        while ct and dtu.running:
            time.sleep(0.5)
            ct -= 1


if __name__ == '__main__':
    test()



# -*- coding: UTF-8 -*
from __future__ import print_function
from xiot import App, MQTTDTU, DTU, __version__
import time

try:
    from queue import Queue, Empty
except ImportError:  # python < 3.0
    from Queue import Queue, Empty
from threading import Thread


class ReadTimeoutException(Exception):
    pass


class Pipe(object):
    """A wrapper around a pipe opened for reading"""

    def __init__(self, pipe):
        self.pipe = pipe
        self.queue = Queue()
        self._runing = True
        self.thread = Thread(target=self._loop)
        self.thread.start()

    def readline(self, timeout=None):
        "A non blocking readline function with a timeout"
        try:
            return self.queue.get(True, timeout)
        except Empty:
            raise ReadTimeoutException()

    def _loop(self):
        try:
            while self._runing:
                line = self.pipe.readline()
                self.queue.put(line)
        except (ValueError, IOError):  # pipe was closed
            pass

    def close(self):
        self._runing = False
        self.pipe.close()


class ShellDTU(DTU):
    def __init__(self, shell):
        self.shell = shell
        DTU.__init__(self, app=self.shell.app, dtu_id=self.shell.gw_id)

    def on_action(self, dev_id, action, value=None):
        self.shell.on_action(dev_id, action, value)

    def on_notify(self, dev_id, data):
        self.shell.on_notify(dev_id, data)


class ShellMQTTDTU(MQTTDTU):
    def __init__(self, shell):
        self.shell = shell
        MQTTDTU.__init__(self, app=self.shell.app, dtu_id=self.shell.gw_id,
                         mqtt_host=self.shell.mqtt_host,
                         mqtt_port=self.shell.mqtt_port,
                         mqtt_username=self.shell.mqtt_username,
                         mqtt_password=self.shell.mqtt_password)

    def on_action(self, dev_id, action, value=None):
        return self.shell.on_action(dev_id, action, value)

    def on_notify(self, dev_id, data):
        return self.shell.on_notify(dev_id, data)


class Runner(object):
    running = True

    def __init__(self, name=None):
        if not name:
            name = self.__class__.__name__
        self.name = name

    def get_attributes(self):
        pass

    def get_actions(self):
        pass

    def get_charts(self):
        pass

    def get_name(self):
        return self.name

    def __get_type_of_value(self, v):
        return 'float' if isinstance(v, float) else \
            'int' if isinstance(v, int) else \
                'bool' if isinstance(v, bool) else \
                    'text' if isinstance(v, type('')) or isinstance(v, type(u'')) else ''

    def get_define(self):
        attributes = self.get_attributes() or []
        for atd in attributes:
            if 'type' not in atd:
                v = atd['value']
                atd['type'] = self.__get_type_of_value(v)
            if 'name' not in atd:
                atd['name'] = atd['attr_id']
        define = {
            'name': self.get_name(),
            'attributes': attributes or [],
            'charts': self.get_charts() or [],
            'actions': self.get_actions() or [],
        }
        return {
            k: v for k, v in define.items() if v != None
        }

    def on_action(self, dtu, dev_id, action, value=None):
        pass

    def start_thread(self, target, args=()):
        th = Thread(target=target, args=args)
        th.setDaemon(True)
        th.start()
        return th

    def stop(self):
        self.running = False

    def get_time(self):
        return 10


class Shell(object):
    def __init__(self, servurl, appid, appkey, mqtturi, gateway, device, runner, name=None, interval=30):
        try:
            import urlparse as parse
        except:
            from urllib import parse
        if mqtturi:
            res = parse.urlparse(mqtturi)
            if res.scheme != 'mqtt':
                raise Exception(u'Unsupported URI: %s' % mqtturi)
            self.mqtt_host = res.hostname
            self.mqtt_port = int(res.port or 1883)
            self.mqtt_username = res.username
            self.mqtt_password = res.password
        else:
            self.mqtt_host = None
        self.app = App(servurl, appid, appkey)
        self.gw_id = gateway
        self.dev_id = device
        self.runner = runner
        self.name = name
        interval = int(interval)
        if interval <= 0:
            interval = 30
        self.interval = interval
        self.running = False

    def new_dtu(self):
        if self.mqtt_host:
            dtu = ShellMQTTDTU(self)
        else:
            dtu = ShellDTU(self)
        return dtu

    def on_ticks(self):
        pass

    def get_default_attributes(self):
        return [
            {'attr_id': 'version', 'name': '版本', 'value': __version__},
            {'attr_id': 'type', 'name': '类型', 'value': 'shell'},
            {'attr_id': 'code', 'name': '代码', 'value': self.dev_id},
        ]

    def __shell_run__(self):
        while True:
            dtu = None
            while self.running:
                try:
                    try:
                        if dtu:
                            print('stopping old dtu')
                            dtu.stop()
                            print('stopped old dtu ok')
                            dtu = None
                    except:
                        pass
                    dtu = self.new_dtu()
                    self.dtu = dtu
                    define = self.runner.get_define()
                    attrs = define.get('attributes') or []
                    attrs += self.get_default_attributes() or []
                    if self.name:
                        define['name'] = self.name
                    define['attributes'] = attrs
                    print('registering %s ...' % define['name'])
                    self.app.doApi('register', {'dtu_id': self.dtu.dtu_id, 'dev_id': self.dev_id, 'define': define})
                    time.sleep(1)
                    print('running...')
                    lastdata = {}

                    def update():
                        if not lastdata:
                            vals = attrs
                        else:
                            vals = self.runner.get_attributes()
                        if not vals:
                            return
                        nd = {
                            d['attr_id']: d['value'] for d in vals if d['value'] != lastdata.get(d['attr_id'])
                        }
                        if nd:
                            # print('update', nd)
                            dtu.updateValues(self.dev_id, nd)
                        lastdata.update(nd)

                    while self.running:
                        update()
                        self.on_ticks()
                        ct = self.interval
                        while ct > 0 and self.running:
                            time.sleep(1)
                            ct -= 1
                except:
                    import traceback
                    traceback.print_exc()
                    time.sleep(3)
            if dtu:
                dtu.sendEvent(self.dev_id, 'normal', u'主动断开连接')
                dtu.updateValues(self.dev_id, {'status': 'disconnected'})
                time.sleep(3)
                dtu.stop()
                self.dtu = None
        self.running = False

    def start_thread(self, target, args=()):
        th = Thread(target=target, args=args)
        th.setDaemon(True)
        th.start()
        return th

    def start(self):
        if self.running:
            return
        self.running = True
        self.start_thread(self.__shell_run__)

    def stop(self):
        self.running = False
        self.runner.stop()

    def on_notify(self, dev_id, data):
        '''通知'''
        # print('shell on_notify', dev_id, data)
        pass

    def on_action(self, dev_id, action, value):
        if dev_id == self.dev_id:
            return self.runner.on_action(self.dtu, dev_id, action, value)


class CommandRunner(Runner):
    def __init__(self, name=None, commands=[]):
        Runner.__init__(self, name)
        cmdmap = {}
        cmds = []
        for arg in commands or []:
            if not arg:
                continue
            if isinstance(arg, dict):
                cmdd = arg
            else:
                if '=' not in arg:
                    print('error command: %s' % arg)
                    exit(-1)
                ix = arg.index('=')
                act_id = arg[0:ix]
                cmdd = {
                    'act_id': act_id,
                    'cmd': arg[ix + 1:]
                }
            cmdmap[cmdd['act_id']] = cmdd
            cmds.append(cmdd)
        self.cmdmap = cmdmap
        self.commands = cmds

    def get_actions(self):
        return [
            {
                "name": d.get('name', d['act_id']),
                "act_id": d['act_id'],
                "type": "button",
            } for d in self.commands
        ]

    def on_action(self, dtu, dev_id, action, value=None):
        if action in self.cmdmap:
            self.run_cmd_action(dtu, dev_id, action, self.cmdmap[action], value)
        else:
            print('unknown acction', action)

    def run_cmd_action(self, dtu, dev_id, action, cmdd, value=None):
        import subprocess, os, uuid
        cmd = cmdd['cmd']
        if cmdd.get('shell'):
            args = cmd
        else:
            args = cmd.split(' ')
        sid = str(uuid.uuid4())
        needret = cmdd.get('return', False)

        def run():
            import time, socket, os
            print(action, args)
            dtu.sendEvent(dev_id, 'warning', 'start %s' % action)
            outs = []
            mid = socket.gethostname()

            def notify(finish=False):
                if needret:
                    return
                text = ''.join(outs)
                dtu.sendNotify(dev_id, {'type': 'result', 'text': text, 'value': value, 'finish': finish, 'sid': sid, 'mid': mid})
                del outs[:]

            try:
                env = value.get('env') if isinstance(value, dict) else None
                if env:
                    env.update(os.environ)
                # subprocess.call()
                p = subprocess.Popen(args, shell=cmdd.get('shell'), stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=os.getcwd(), bufsize=1, env=env)
                pipe = Pipe(p.stdout)
                last = time.time()
                start = time.time()
                outs.append('\n>run start...\n\n')
                notify()
                while self.running:
                    # notify()
                    try:
                        out = pipe.readline(1)
                        if out:
                            outs.append(out.decode('utf-8'))
                        elif p.poll() != None:
                            break
                    except ReadTimeoutException as e:
                        # import traceback
                        # traceback.print_exc()
                        pass
                    timeoff = (time.time() - last)
                    if timeoff > 1 and outs or timeoff > 5:
                        last = time.time()
                        notify()
                # print("self.running out", self.running, p.returncode)
                code = p.returncode
                ctime = int((time.time() - start) * 1000)
                outs.append('\n>run end, exit code=%s, time %dms!\n' % (code, ctime))
                notify(True)
                print('%s exit code=%s, time %dms!' % (action, code, ctime))
                try:
                    pipe.close()
                except:
                    pass
                if needret and outs:
                    dtu.sendEvent(dev_id, 'normal', ''.join(outs))
                    time.sleep(.5)
                dtu.sendEvent(dev_id, 'success', 'finish %s' % action)
            except Exception as e:
                import traceback
                traceback.print_exc()
                dtu.sendEvent(dev_id, 'error', str(e))
                try:
                    outs.append('\nerror!!!\n')
                    outs.append(str(e))
                    outs.append('\nrun end, time %dms!\n' % (int((time.time() - start) * 1000)))
                    notify(True)
                except:
                    pass

        self.start_thread(run)


class SysInfoRunner(CommandRunner):
    def __init__(self, name=None, commands=None, monitor=True):
        self.monitor = monitor
        CommandRunner.__init__(self, name=name, commands=commands)

    def get_attributes(self):
        if self.monitor:
            import psutil, datetime
            return [
                {
                    'attr_id': 'now',
                    'value': datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                    'name': '时间',
                    'unit': ''
                },
                {
                    'attr_id': 'cpu',
                    'value': psutil.cpu_percent(),
                    'name': 'CPU',
                    'unit': '%'
                },
                {
                    'attr_id': 'mem',
                    'value': psutil.virtual_memory().percent,
                    'name': '内存',
                    'unit': '%'
                },
                {
                    'attr_id': 'disk',
                    'value': psutil.disk_usage('/').percent,
                    'name': '磁盘',
                    'unit': '%'
                }
            ]

    def get_charts(self):
        if self.monitor:
            return [
                {
                    "name": "系统信息",
                    "type": "timeline",
                    "size": "large",
                    "fields": "cpu,mem,disk",
                },
            ]


def sys_exit(code):
    try:
        exit(code)
    except:
        pass


if __name__ == '__main__':
    import argparse, json

    parser = argparse.ArgumentParser(add_help=True)
    parser.add_argument('-s', '--server', help='The Server URL', type=str, required=False)
    parser.add_argument('-m', '--mqtt', help='The MQTT Uri, Example: mqtt://username:password@xxx.com:1883', type=str, required=False)
    parser.add_argument('-a', '--appid', help='The xiot appid', type=str, required=False)
    parser.add_argument('-k', '--appkey', help='The xiot appkey', type=str, required=False)
    parser.add_argument('-g', '--gateway', help='The Gateway ID', type=str, required=False)
    parser.add_argument('-d', '--device', help='The Device ID', type=str, required=False)
    parser.add_argument('-n', '--name', help='The Device Name', type=str, required=False)
    parser.add_argument('-i', '--interval', help='The Update Interval(seconds)', default=0, type=int, required=False)
    parser.add_argument('-c', '--command', action='append', type=str, help='The Command, Example: list="echo hi"')
    parser.add_argument('--config', type=str, help='The Config File/URI, Example: define.yaml')
    parser.add_argument('--runner', type=str, help='The Define JSON String')
    args = parser.parse_args()
    if args.runner:
        args.runner = json.loads(args.runner)
    if args.config:
        # define
        import os, socket

        rawargs = args.__dict__  # {k: v for k, v in args.__dict__.items()}
        uri = args.config
        if '://' in uri:
            import requests

            print('fetching config...')
            txt = requests.get(uri).text
        else:
            import io, os

            if not os.path.exists(uri):
                raise Exception('define file not exists: %s' % uri)
            txt = io.open(uri, encoding='utf8').read()
        cxt = {'pid': os.getpid(), 'hostname': socket.gethostname()}
        for k, v in cxt.items():
            if not v:
                continue
            txt = txt.replace('${%s}' % k, str(v))
        if '.yaml' in uri:
            import yaml

            newargs = yaml.safe_load(txt)
        elif '.json' in uri:
            newargs = json.loads(txt)
        else:
            raise Exception(u'unknow define type: %s' % uri)
        for k, v in newargs.items():
            if not rawargs.get(k):
                rawargs[k] = v
    sh = None

    try:
        import socket

        runner_clz_map = {
            c.__name__: c for c in [CommandRunner, SysInfoRunner]
        }
        if args.runner:
            runner_clz = runner_clz_map.get(args.runner['class'])
            if not runner_clz:
                sclz = args.runner['class']
                rix = sclz.rindex('.')
                p = __import__(sclz[0:rix])
                runner_clz = getattr(p, sclz[rix + 1:])
            runner_args = args.runner.get('args', {})
        elif args.command:
            runner_clz = CommandRunner
            runner_args = {'commands': args.command}
        # print(runner_clz, runner_args)
        runner = runner_clz(**runner_args)
        sh = Shell(servurl=args.server, appid=args.appid, appkey=args.appkey, mqtturi=args.mqtt, gateway=args.gateway, device=args.device, runner=runner, name=args.name, interval=int(args.interval))
        sh.start()
        while sh.running:
            time.sleep(1)
            # print('sleepping...')
    except KeyboardInterrupt:
        if sh:
            print('exit...')
            sh.stop()
            time.sleep(3)
        sys_exit(0)



from setuptools import setup

from xiot import __version__

setup(
    name='xiot',
    keywords=('xiot'),
    version=__version__,
    py_modules=['xiot', 'xiotshell'],
    url='https://github.com/sintrb/xiot',
    license='Apache License 2.0',
    author='Robin',
    author_email='sintrb@gmail.com',
    description='A lib for xiot.',
    install_requires=['requests', 'paho-mqtt', 'six'],
)
